/*
 * This macro allows you to use a 'C' function to handle an AVR interrupt.
 *
 * The only parameter is the DECIMAL **ADDRESS** of the interrupt vector.
 *
 * The macro MUST be placed IMMEDIATELY preceeding the definition of the
 * function which is to be the interrupt handler. Example:
 *
 * 		INTERRUPT(_INT0_) serial_int_handler()
 * 		{
 *			/* ... Function code to handle serial interrupt ... */
 *		}
 *
 *
 *NOTE1: The parameter to this macro is the MEMORY ADDRESS (not the vector
 * number) of the interrupt vector in *DECIMAL* format. Also note that the
 * DDS assembler uses BYTE addressing (not word addressing as shown in the
 * AVR data book)... This means that INT0=2, INT1=4, TIMER1 CAPT=6 etc.
 *
 *NOTE2: By default, registers are saved on the SOFTWARE(Y) stack, which
 * usually has the most room. If you prefer to use the HARDWARE stack,
 * #define INT_HSTACK before you #include this file. *** The interrupt
 * always saves the return address (2 bytes) on the hardware stack, and
 * the registers will add 11 more... MAKE SURE THE STACK IS LARGE ENOUGH!
 *
 *NOTE3: In the interest of efficency, this macro saves only the basic AVR
 * registers used by the compiler generated code and the runtime support files.
 * If you call any library functions within your interrupt handler, you must
 * make sure that they do not corrupt any other registers...
 *
 *NOTE4: The 'Y' (R28:R29) register is not saved/preserved, because it is
 * always maintained as a stack pointer. You may use space on this stack,
 * however the R28:R29 pointer must be left the same on exit as on entry.
 *
 *NOTE5: Make sure there is adaquate space reserved in AVR_RLP.ASM for the
 * interrupt vector you are planning to use. Should look something like:
 *		 ORG	$0000		; Code goes here
 *		 RJMP	?begin		; Launch startup code
 *		 DS		12*2		; Reserve room for interrupt vectors
 *; Initialize process - Setup stack pointers
 *?begin LDI	R24,?HSTACK	; Get low hardware stack
 *		 ...
 *
 * This file **REQUIRES** the extended pre-processor (MCP).
 * (CCAVR ... -P, or "Preprocess" step enabled in DDSIDE)
 *
 * Copyright 1998-2005 Dave Dunfield
 * All rights reserved.
 */
#ifdef INT_HSTACK
 #define push(r) PUSH r		/* Push on hardware stack */
 #define pop(r)	POP	r		/* Pop from hardware stack */
#else
 #define _					/* Used to force embedded space */
 #define push(r) ST _-Y,r	/* Push on software stack */
 #define pop(r) LD r,Y+		/* Pop from software stack */
 #undef _					/* No longer needed */
#endif

#define	INTERRUPT(vec) asm {\
$SE:0						/* Insure output in code segment */\
_/**/vec	/* This label will be the address of the "prefix" stub */\
		ORG		vec			/* Position to interrupt vector */\
		RJMP	_/**/vec	/* Enter the label address */\
		ORG		_/**/vec	/* Position back to original PC */\
		push	(R0)		/* Save R0 */\
		IN		R0,SREG-$20	/* Get condition codes */\
		push	(R0)		/* Save condition codes */\
		push	(R19)		/* Save remaining registers */\
		push	(R22)\
		push	(R23)\
		push	(R24)\
		push	(R25)\
		push	(R26)\
		push	(R27)\
		push	(R30)\
		push	(R31)\
		RCALL	__/**/vec	/* Execute user handler */\
		pop		(R31)		/* Restore registers */\
		pop		(R30)\
		pop		(R27)\
		pop		(R26)\
		pop		(R25)\
		pop		(R24)\
		pop		(R23)\
		pop		(R22)\
		pop		(R19)\
		pop		(R0)		/* Restore condition codes */\
		OUT		SREG-$20,R0	/* Reset condition codes */\
		pop		(R0)		/* Restore R0 */\
		RETI				/* End interrupt */\
__/**/vec	/* This label addresses the function */\
}
#undef push					/* No longer needed */
#undef pop					/* No longer needed */

/* Common AVR interrupt vector addresses */
#define _INT0_			2	/* External interrupt request 0 */
#define	_INT1_			4	/* External interrupt request 1 */
#define	_TIMER1CAPT_	6	/* Timer/Counter1 capture event */
#define	_TIMER1COMPA_	8	/* Timer/Counter1 compare match A */
#define	_TIMER1COMPB_	10	/* Timer/Counter1 compare match B */
#define	_TIMER1OVF_		12	/* Timer/Counter1 overflow */
#define	_TIMER0OVF_		14	/* Timer/Counter0 overflow */
#define	_SPISTC_		16	/* Serial transfer complete */
#define	_UARTRX_		18	/* UART RX complete */
#define	_UARTUDRE_		20	/* UART data register empty */
#define	_UARTTX_		22	/* UART TX complete */
#define	_ANACOMP_		24	/* Analog comparator */
